/*
	USB-PDC(AN2131SC) JTAG Probe Interface DLL Version 1.0

	Copyright(c) 2002 FREE WING (AN2131SC porting)
	Copyright(c) 2002 Tatsuyuki Satoh   (base Driver & DDL conversion)
	Copyright(c) 2002 Optimize (EZ-USB firmware & common driver)

	2001. 8.05 1st release

	support probe
		USB-PDC or Cypress AN2131SC(EZ-USB) without EEPROM
		+ Cypress EZ-USB DevKit(Device Driver)

	I/O port assign for JTAG mode

		PC5(13)  -> TCK
		PC1(6)   -> TDI
		PC4(12)  -> TMS
		PC0(7)   <- TDO
		PC2(15)  <- Target VCC 3.3V

	Export DLL Functions

	JTAG BounteryScan
		DWORD jtag_version(void)
		int jtag_init(DWORD port_address,DWORD baudrate,char *probe_name)
		int jtag_power(void)
		int jtag_txrx(const BYTE *ibuf,BYTE *obuf,DWORD drsize,int tms_s,int tms_e)
		void jtag_end(void)

*/
#include <windows.h>
#include <stdio.h>

// DLL Export prefix
#define DLLEXPORT __declspec(dllexport)
#ifdef __MINGW32__
#define DLLAPI
#else
#define DLLAPI WINAPI
#endif

// Warning fix for EZ-USB header file with MINGW
#if !defined(_BYTE_DEFINED)
#define _BYTE_DEFINED
#define _WORD_DEFINED
#endif

// EZ-USN & chamereonUSB General-purpose FirmWare */
#include "cusb.h"
#include "gpfw_ex.h"

/* ---------- Hardware pin assign ---------- */

//		TCK			FWR
#define CF_OEC			0x3A			/* PORT-C OE INIT */
#define CF_TDO(port)	((port)&1)		/* PC0 */
#define CF_TDI			0x02			/* PC1 */
#define CF_VCC(port)	((port)&4)		/* PC2 */
#define CF_TMS			0x10			/* PC4 */
#define CF_TCK			0x20			/* PC5 */

//#define CF_PROGRAM    0x01       /* PC0 */
#define CF_PROGRAM    0x20       /* PC6 */
#define CF_DIN        0x01       /* PC0 */
#define CF_DONE(port) ((port)&4) /* PC3 */

// Pxɓ]őoCg
#define TX_MAX 65536*4 //(2048*1024)
#define TXRX_MAX 1

// handle for EZ-USB
static HANDLE usb_h = NULL;

// USB command buffer
static u8 cmd_buf[64];

// USB FirstWrite buffer
static u8 usb_buf[TX_MAX];

/* -------------------- chamereonUSB General-purpose FirmWare interface -------------------- */

// Write to CPIPE (command)
static void usb_cpipe_write(void *buf,DWORD size)
{
	usb_bulk_write(&usb_h,GPFW_CPIPE,buf,size);
}

// Read from RPIPE (read data)
static void usb_rpipe_read(void *buf,DWORD size)
{
	usb_bulk_read(&usb_h,GPFW_RPIPE,buf,size);
}

// Fast WRTITE Transfer
static void usb_fwrite(u8 gpio,void *data,DWORD size)
{
	u8 cmd[4];

	cmd[0]=GPFW_BWRITE_FW | (gpio&0x0f);
	cmd[1]=(u8)(size & 0x3f);
	cmd[2]=(u8)(size >> 6);
	cmd[3]=(u8)(size >> 14);
	usb_bulk_write(&usb_h,GPFW_CPIPE,cmd,4);
	usb_bulk_write(&usb_h,GPFW_WPIPE,data,size);
}

// Initialize
static int ezusb_init(int no)
{
	// open EZ-USB & download chanereon usb firmware
	printf("Setup EZ-USB with chamereonUSB GP firnware V1.10 for USB-PDC Ver0.00\n");
	if(cusb_init(no,&usb_h,fw_bin,"GPFW","V11P"))
	{
		usb_h = NULL;
		printf("Can't found EZ-USB\n");
		return 2;
	}
	return 0;
}

/* -------------------- JTAG contyroll -------------------- */
DLLEXPORT DWORD DLLAPI jtag_version(void)
{
	DWORD mjr = 1;
	DWORD mnr = 0;
	return (mjr<<16) + mnr;
}

DLLEXPORT int DLLAPI jtag_init(DWORD port_address,DWORD baudrate,char *probe_name)
{
	// name set
	if( probe_name != NULL)
	{
		sprintf(probe_name,"EZ-USB No.%d : USB-PDC(AN2131SC) (PC5=TCK,PC1=TDI,PC4=TMS,PC1=PC0,PC2=TVCC)"
			,(int)port_address);
	}

	// initialize EZ-USB
	if( ezusb_init(port_address) )
		return 2;

	// setup GPIO & read Target VCC port
	cmd_buf[0]=GPFW_DIR_FW;
	cmd_buf[1]=CF_OEC;
	cmd_buf[2]=GPFW_SET_FW;
	cmd_buf[3]=CF_TCK;
	usb_cpipe_write(cmd_buf,4); // init & sense

	cmd_buf[0]=GPFW_BUS;  // streach=0
	cmd_buf[1]= 0x12;     // RMOD = 10 , TMOD = 10
	usb_cpipe_write(cmd_buf,2); // init & sense

	return 0;
}

DLLEXPORT int DLLAPI jtag_power(void)
{
	// setup GPIO & read Target VCC port
	cmd_buf[0]=GPFW_GET_FW;
	usb_cpipe_write(cmd_buf,1);
	usb_rpipe_read(cmd_buf,1);

	//Target VCC`FbN
	return CF_VCC(cmd_buf[0])? 1 : 0;
}

DLLEXPORT int DLLAPI jtag_txrx(const BYTE *ibuf,BYTE *obuf,DWORD drsize,int tms_s,int tms_e)
{
	DWORD c;
	DWORD pos;
	DWORD block_pos;
	DWORD block_size;

/*
			printf("[");
	if(obuf) {
			printf("O");
	} else {
			printf("_");
	}
	if(ibuf) {
			printf("I");
	} else {
			printf("_");
	}
*/
	// USB write
	if(obuf == NULL)
	{
//			printf("}");
		// TMS̒lobt@ɍ쐬
		memset(usb_buf,tms_s ? CF_TMS : 0x00,drsize-1);
		usb_buf[drsize-1] = tms_e ? CF_TMS : 0x00;
		// TDI̒lobt@ɍ쐬
		if(ibuf)
		{
			for(pos=0 ; pos<drsize ; pos+=8)
			{
				BYTE value = ibuf[pos/8];
				// nibble -> bitϊ , obt@쐬
				if(value&0x80) usb_buf[pos+0] |= CF_TDI;
				if(value&0x40) usb_buf[pos+1] |= CF_TDI;
				if(value&0x20) usb_buf[pos+2] |= CF_TDI;
				if(value&0x10) usb_buf[pos+3] |= CF_TDI;
				if(value&0x08) usb_buf[pos+4] |= CF_TDI;
				if(value&0x04) usb_buf[pos+5] |= CF_TDI;
				if(value&0x02) usb_buf[pos+6] |= CF_TDI;
				if(value&0x01) usb_buf[pos+7] |= CF_TDI;
			}
		}
		//M
		usb_fwrite(0,usb_buf,drsize);
	}
	else
	{
//			printf("]");
		//clear outbuf
		if(obuf)
			for(c=0;c<(drsize+7)/8;c++)
				obuf[c] = 0;

		block_size	= 0;
		for(block_pos=0;block_pos<drsize;block_pos+=block_size)
		{
//			printf("[%08lx]", block_pos);

			//̑MTCY̐ݒ
			block_size = (block_pos+TXRX_MAX) < drsize ? TXRX_MAX : drsize-block_pos;
			//Mf[^쐬
			for(c=block_pos;c<block_pos+block_size;c++)
			{
				BYTE tdi=0,tms;

		if(ibuf) {
				tdi = (ibuf[c/8]&(0x80>>(c&7))) ? 1 : 0;
		}
				tms = (c==drsize-1) ? tms_e : tms_s;
		if(obuf) {
				usb_buf[(c-block_pos)*3+0] = GPFW_WRITE_FW;
				usb_buf[(c-block_pos)*3+1] = (tdi ? CF_TDI : 0) | (tms ? CF_TMS : 0);
				usb_buf[(c-block_pos)*3+2] = GPFW_GET_FW;
		} else {
				usb_buf[(c-block_pos)*2+0] = GPFW_WRITE_FW;
				usb_buf[(c-block_pos)*2+1] = (tdi ? CF_TDI : 0) | (tms ? CF_TMS : 0);
		}
			}

		if(obuf) {
			usb_cpipe_write(usb_buf,block_size*3); // TX data
			usb_rpipe_read(usb_buf,block_size);    // read TDO
		} else {
			usb_cpipe_write(usb_buf,block_size*2); // TX data
		}

			// TDOf[^̕ϊ
	if(obuf) {
			for(c=block_pos;c<block_pos+block_size;c++)
			{
				if( CF_TDO(usb_buf[c-block_pos]) )
					obuf[c/8]  |= 0x01<<(c&7);
			}
	}

		}
	}
	return 0;
}

DLLEXPORT void DLLAPI jtag_end(void)
{
	if(usb_h)
		usb_close(&usb_h);
}

